@@ -34,8 +34,10 @@ use CPANPLUS::Backend 0.051;
 use CPANPLUS::Configure;
 use CPANPLUS::Error;
 
+use File::Path;
+use File::Basename;
 use File::HomeDir qw( home );
-use File::Spec::Functions qw( catfile );
+use File::Spec::Functions qw( splitpath catfile );
 use LWP::Simple;
 use POSIX qw( O_CREAT O_RDWR );         # for SDBM_File
 use Regexp::Assemble;
@@ -49,15 +51,16 @@ use Config::IniFiles;
 # use YAML 'Dump';
 
 require Test::Reporter;
+require YAML;
 
-our $VERSION = '0.02_06';
+our $VERSION = '0.03_03';
 $VERSION = eval $VERSION;
 
 require Exporter;
 
 our @ISA = qw( Exporter );
 our %EXPORT_TAGS = (
-  'all'      => [ qw( mark test excluded ) ],
+  'all'      => [ qw( mark test excluded purge flush ) ],
   'default'  => [ qw( mark test excluded ) ],
 );
 
@@ -71,6 +74,11 @@ use constant CONFIG_FILE   => 'cpansmoke.ini';
 
 my $extn = qr/(?:\.(?:tar\.gz|tgz|zip))/;	# supported archive extensions
 
+=head1 OBJECT INTERFACE
+
+=over 4
+
+=cut
 
 {
   my %Checked;
@@ -80,11 +88,11 @@ my $extn = qr/(?:\.(?:tar\.gz|tgz|zip))/;	# supported archive extensions
   # function creates an additional reference in the scope of an
   # if (tied %Checked) { ... } which causes a warning etc.
 
-  sub connect_db {
+  sub _connect_db {
     my $self = shift;
-    my $filename = shift || catfile($self->basedir(), DATABASE_FILE);
+    my $filename = $self->{database_file};
     if ($TiedObj) {
-      # error("Already connected to the database!");
+#     error("Already connected to the database!");
     } else {
       $TiedObj = tie %Checked, 'SDBM_File', $filename, O_CREAT|O_RDWR, 0644;
       $self->{checked} = \%Checked;
@@ -92,7 +100,7 @@ my $extn = qr/(?:\.(?:tar\.gz|tgz|zip))/;	# supported archive extensions
     }
   }
 
-  sub disconnect_db {
+  sub _disconnect_db {
     my $self = shift;
 
     if ($TiedObj) {
@@ -100,80 +108,85 @@ my $extn = qr/(?:\.(?:tar\.gz|tgz|zip))/;	# supported archive extensions
       $self->{checked} = undef;
       untie %Checked;
       $self->_debug("Disconnected from database.");
-      #		} else {
-      #			error("Not connected to the database!");
+#	} else {
+#	  error("Not connected to the database!");
     }
   }
 
   my $CONF = CPANPLUS::Configure->new();
-  sub connect_configure {
+  sub _connect_configure {
     return $CONF;
   }
 
   my $CpanPlus;
 
-  sub connect_cpanplus {
+  sub _connect_cpanplus {
     my $self = shift;
     return $self->{cpan} = $CpanPlus if ($CpanPlus);
 
-    my $re = new Regexp::Assemble;
-    $re->add( @{ $self->{exclude_dists} } );
-
     $CpanPlus = CPANPLUS::Backend->new();
 
     if ($CPANPLUS::Backend::VERSION >= 0.052) {
 
+      # TODO: if PASS included skipped tests, add a comment
+
       $CpanPlus->_register_callback(
         name => 'munge_test_report',
         code => sub {
-	  my $mod    = shift;
-	  my $report = shift || "";
-	  $report .=
-	    "\nThis report was machine-generated by CPAN::YACSmoke $VERSION.\n";
-	  return $report;
+		  my $mod    = shift;
+		  my $report = shift || "";
+		  $report .=
+			"\nThis report was machine-generated by CPAN::YACSmoke $VERSION.\n";
+		  return $report;
         },
       );
     }
 
+    # BUG: this callback does not seem to get called consistently, if at all.
+
     $CpanPlus->_register_callback(
       name => 'install_prerequisite',
       code => sub {
-	my $mod   = shift;
-	my $root;
-	if ($mod->package =~ /^(.+)$extn$/) {
-	  $root = $1;
-	}
-	else {
-	  error("Cannot handle ".$mod->package);
-	  return;
-	}
+		my $mod   = shift;
+		my $root;
+		if ($mod->package =~ /^(.+)$extn$/) {
+		  $root = $1;
+		}
+		else {
+		  error("Cannot handle ".$mod->package);
+		  return;
+		}
 
-#	print Dump($obj); # (*)
-	unless ($TiedObj) {
-	  exit error("Not connected to database!");
-	}
-	while (my $arg = shift) {
-	  $arg->package =~ m/^(.+)$extn$/;
-	  my $package = $1;
+		unless ($TiedObj) {
+		  croak "Not connected to database!";
+		}
+		while (my $arg = shift) {
+		  $arg->package =~ m/^(.+)$extn$/;
+		  my $package = $1;
 
-	  if ($package =~ $re->re) { # prereq on excluded list
-	    return;
-	  }
+		  # BUG: Exclusion does not seem to work for prereqs.
+		  # Sometimes it seems that the install_prerequisite
+		  # callback is not even called! Need to investigate.
 
-	  my $checked = $Checked{$package};
-	  if (defined $checked &&
-	      $checked =~ /aborted|fail|unknown|na|ungraded/ ) {
+		  if ($self->_is_excluded_dist($package)) { # prereq on excluded list
+			msg("Prereq $package is excluded");
+			return;
+		  }
 
-	    if ($self->{ignore_bad_prereqs}) {
-	      msg("Known uninstallable prereqs $package - may have problems\n");
-	    } else {
-	      msg("Known uninstallable prereqs $package - aborting install\n");
-	      $Checked{$root} = "aborted";
-	      return;
-	    }
-	  }
-	}
-	return 1;
+		  my $checked = $Checked{$package};
+		  if (defined $checked &&
+			  $checked =~ /aborted|fail|unknown|na|ungraded/ ) {
+
+			if ($self->{ignore_bad_prereqs}) {
+			  msg("Known uninstallable prereqs $package - may have problems\n");
+			} else {
+			  msg("Known uninstallable prereqs $package - aborting install\n");
+			  $Checked{$root} = "aborted";
+			  return;
+			}
+		  }
+		}
+		return 1;
       },
     );
 
@@ -181,34 +194,34 @@ my $extn = qr/(?:\.(?:tar\.gz|tgz|zip))/;	# supported archive extensions
       name => 'send_test_report',
       code => sub {
 
-	unless ($TiedObj) {
-	  exit error("Not connected to database!");
-	}
-	my $mod   = shift;
-	my $grade = lc shift;
-	if ($mod->{package} =~ /^(.+)$extn$/) {
-	  my $package = $1;
-	  my $checked = $Checked{$package};
+		unless ($TiedObj) {
+		  exit error("Not connected to database!");
+		}
+		my $mod   = shift;
+		my $grade = lc shift;
+		if ($mod->{package} =~ /^(.+)$extn$/) {
+		  my $package = $1;
+		  my $checked = $Checked{$package};
 
-	  # TODO: option to report only passing tests
+		  # TODO: option to report only passing tests
 
-	  return unless ($self->{cpantest});
+		  return unless ($self->{cpantest});
 
-	  return if (defined $checked && (
-                      ($checked eq 'aborted' &&  $grade ne 'pass')     ||
-		      ($checked eq 'unknown'  && $grade eq 'unknown')  ||
-		      ($checked eq 'ungraded' && $grade eq 'fail')     ||
-		      ($checked =~ /pass|na/)                          ||
-		      ($checked eq 'fail' && $grade =~ /unknown|na|fail/)));
+		  return if (defined $checked && (
+				  ($checked eq 'aborted' &&  $grade ne 'pass')     ||
+				  ($checked eq 'unknown'  && $grade eq 'unknown')  ||
+				  ($checked eq 'ungraded' && $grade eq 'fail')     ||
+				  ($checked =~ /pass|na/)                          ||
+				  ($checked eq 'fail' && $grade =~ /unknown|na|fail/)));
 
-	  $Checked{$package} = $grade;
+		  $Checked{$package} = $grade;
 
-	  return ((!$self->{report_pass_only}) || ($grade eq 'pass'));
+		  return ((!$self->{report_pass_only}) || ($grade eq 'pass'));
 
-	} else {
-	  error("Unable to parse package information\n");
-	  return;
-	}
+		} else {
+		  error("Unable to parse package information\n");
+		  return;
+		}
       },
     );
 
@@ -225,7 +238,8 @@ my @CONFIG_FIELDS = qw(
 	verbose debug force cpantest
 	recent_list_age ignore_cpanplus_bugs fail_max
 	exclude_dists test_max audit_log
-        ignore_bad_prereqs report_pass_only
+	ignore_bad_prereqs report_pass_only
+	allow_retries flush_flag
 );
 
 my @CPANPLUS_FIELDS = qw(
@@ -235,10 +249,6 @@ my @CPANPLUS_FIELDS = qw(
 );
 
 
-=head1 OBJECT INTERFACE
-
-=over 4
-
 =item new( [ %config ] )
 
 The object interface is created normally through the test() or mark()
@@ -259,18 +269,21 @@ Configuration settings are:
   fail_max
   exclude_dists
   test_max
+  allow_retries
+  flush_flag         - used by purge()
 
   list_from          - List plugin required, default Recent
 
   recent_list_age    - used with the Recent plugin 
   recent_list_path   - used with the Recent plugin 
   mailbox            - used with the Outlook plugin 
-  nntp_id            - used with the NNTP plugin 
+  nntp_id            - used with the NNTP plugins 
   webpath            - used with the WebList plugin 
 
   audit_log          - log file to write progress to
 
   config_file        - an INI file with the above settings
+  database_file      - the local cpansmoke database
 
 All settings can use defaults. With regards to the last setting,
 the INI file should contain one setting per line, except the values
@@ -291,16 +304,27 @@ tested in a single run. As some distributions can take some time to be
 tested, it may be more suitable to run in small batches at a time. The
 default setting is 100 distributions.
 
+The setting 'allow_retries' defaults to include grades of NONE, UNGRADED
+and ABORTED. If you wish to change this, for example to only allow grades
+of NONE and UNGRADED to be retried, you can specify as:
+
+  [CONFIG]
+  allow_retries=none|ungraded
+
 =back
 
 =cut 
 
 sub new {
 	my $class = shift || __PACKAGE__;
-	## Ensure CPANPLUS knows we automated
-	$ENV{AUTOMATED_TESTING} = 1;
 
-	my $conf = connect_configure();
+	## Ensure CPANPLUS knows we automated. 
+	## (Q: Should we use Env::C to set this instead?)
+
+	$ENV{AUTOMATED_TESTING} = 1;
+	$ENV{PERL_MM_USE_DEFAULT} = 1; # despite verbose setting
+
+	my $conf = _connect_configure();
 
 	## set internal defaults
 	my $self  = {
@@ -310,6 +334,7 @@ sub new {
 		fail_max             => 3,     # max failed versions to try
 		exclude_dists        => [ ],   # Regexps to exclude
 		test_max             => 100,   # max distributions per run
+		allow_retries		 => 'none|ungraded|aborted',
 	};
 
 	bless $self, $class;
@@ -319,9 +344,10 @@ sub new {
 		$self->{$field} = $conf->get_conf($field) || 0;
 	}
 
+
 	## force overide of default settings
 	$self->{skiptest} = 0;
-	$self->{prereqs}  = 1;
+	$self->{prereqs}  = 2; # force to ask callback
 	
 	my %config = @_;
 
@@ -339,6 +365,14 @@ sub new {
 		$self->{exclude_dists} = [ @list ]	if(@list);
 	}
 
+	if ($self->{audit_log}) {
+	  my ($vol, $path, $file) = splitpath $self->{audit_log};
+	  unless ($vol || $path) {
+	    $self->{audit_log} = catfile($self->basedir(), $file);
+	  }
+	}
+
+
 	## command line switches override
 	foreach my $field (@CONFIG_FIELDS, 'audit_cb') {
 		if (exists $config{$field}) {
@@ -369,16 +403,22 @@ sub new {
 	$config{smoke} = $self;
 	$self->{plugin} = $plugin->new(\%config);
 
-	$self->connect_db();
-	$self->connect_cpanplus();
+
+	## determine the database file
+
+	$self->{database_file} ||= catfile($self->basedir(), DATABASE_FILE);
+
+	$self->_connect_db();
+	$self->_connect_cpanplus();
 
 	return $self;
 }
 
 
 sub DESTROY {
-	my $self = shift;
-	$self->disconnect_db();
+  my $self = shift;
+  $self->_audit("Disconnecting from database");
+  $self->_disconnect_db();
 }
 
 =head2 METHODS
@@ -394,38 +434,70 @@ Obtains the users home directory
 # TODO: use CPANPLUS function
 
 sub homedir {
-	my $self = shift;
-	return $self->{homedir} = shift	if (@_);
+  my $self = shift;
+  return $self->{homedir} = shift	if (@_);
 
-	unless (defined $self->{homedir}) {
-		if ($^O eq "MSWin32") { # bug in File::HomeDir <= 0.06
-			$self->{homedir} =	$ENV{HOME}						 ||
-								($ENV{HOMEDRIVE}.$ENV{HOMEPATH}) ||
-								$ENV{USERPROFILE}                ||
-								home();
-		} else {
-			$self->{homedir} = home();
-		}
-	}
-	return $self->{homedir};
+  unless (defined $self->{homedir}) {
+    if ($^O eq "MSWin32") { # bug in File::HomeDir <= 0.06
+      $self->{homedir} = $ENV{HOME}      ||
+	($ENV{HOMEDRIVE}.$ENV{HOMEPATH}) ||
+	  $ENV{USERPROFILE}              ||
+	    home();
+    } else {
+      $self->{homedir} = home();
+    }
+  }
+  $self->_audit("homedir = " . $self->{homedir});
+  return $self->{homedir};
 }
 
 =item basedir
 
 Obtains the base directory for downloading and testing distributions.
 
+=cut 
+
+sub basedir {
+  my $self = shift;
+  return $self->{basedir} = shift if (@_);
+
+  unless (defined $self->{basedir}) {
+    $self->{basedir} = $self->{conf}->get_conf("base") || $self->homedir();
+  }
+  return $self->{basedir};
+}
+
+=item builddir
+
+Obtains the build directory for unpacking and testing distributions.
+
 =back
 
 =cut 
 
-sub basedir {
-	my $self = shift;
-	return $self->{basedir} = shift	if (@_);
+sub builddir {
+  my $self = shift;
 
-	unless (defined $self->{basedir}) {
-		$self->{basedir} = $self->{conf}->get_conf("base") || $self->homedir();
+  require Config;
+
+  return File::Spec->catdir(
+    $self->{conf}->get_conf('base'),
+	$Config::Config{version},
+	$self->{conf}->_get_build('moddir'),
+  );
+}
+
+
+sub _is_excluded_dist {
+	my $self = shift;
+	my $dist = shift;
+	unless($self->{re}) {
+		$self->{re} = new Regexp::Assemble;
+		$self->{re}->add( @{ $self->{exclude_dists} } );
 	}
-    return $self->{basedir};
+
+	return 1	if($dist =~ $self->{re}->re);
+	return 0;
 }
 
 sub _remove_excluded_dists {
@@ -433,12 +505,10 @@ sub _remove_excluded_dists {
 	my @dists = ( );
 	my $removed = 0;
 
-	my $re = new Regexp::Assemble;
-	$re->add( @{ $self->{exclude_dists} } );
-
 	while (my $dist = shift) {
-		if ($dist =~ $re->re) {
-			chomp($dist);
+		my $file = basename($dist);
+		if ($self->_is_excluded_dist($file)) {
+			chomp($file);
 			$self->_track("Excluding $dist");
 			$removed = 1;
 		} else {
@@ -450,49 +520,49 @@ sub _remove_excluded_dists {
 }
 
 sub _build_path_list {
-	my $self = shift;
-	my $ignored = 0;
+  my $self = shift;
+  my $ignored = 0;
 
-	my %paths = ( );
-	while (my $line = shift) {
-		if ($line =~ /^(.*)\-(.+)(\.tar\.gz)$/) {
-			my $dist = $1;
-			my @dirs = split /\/+/, $dist;
-			my $ver  = $2;
+  my %paths = ( );
+  while (my $line = shift) {
+    if ($line =~ /^(.*)\-(.+)$extn$/) {
+      my $dist = $1;
+      my @dirs = split /\/+/, $dist;
+      my $ver  = $2;
 
-			# due to rt.cpan.org bugs #11093, #11125 in CPANPLUS
+      # due to rt.cpan.org bugs #11093, #11125 in CPANPLUS
 
-			if ($self->{ignore_cpanplus_bugs} || (
-				(@dirs == 4) && ($ver =~ /^[\d\.\_]+$/)) ) {
+      if ($self->{ignore_cpanplus_bugs} || (
+	   (@dirs == 4) && ($ver =~ /^[\d\.\_]+$/)) ) {
 
-				if (exists $paths{$dist}) {
-					unshift @{ $paths{$dist} }, $ver;
-				} else {	
-					$paths{$dist} = [ $ver ];
-				}
-
-			} else {
-				$self->_track("Ignoring $dist-$ver (due to CPAN+ bugs)");
-				$ignored = 1;
-			}
-
-		# check for previously parsed package string
-		} elsif($line =~ /^(.*)\-(.+)$/) {
-			my $dist = $1;
-			my @dirs = split /\/+/, $dist;
-			my $ver  = $2;
-
-			if(@dirs == 1) {	# previously parsed
-				if (exists $paths{$dist}) {
-					unshift @{ $paths{$dist} }, $ver;
-				} else {	
-					$paths{$dist} = [ $ver ];
-				}
-			}
+		if (exists $paths{$dist}) {
+		  unshift @{ $paths{$dist} }, $ver;
+		} else {	
+		  $paths{$dist} = [ $ver ];
 		}
-	}
-	$self->_audit('')	if($ignored);
-	return %paths;
+
+      } else {
+		$self->_track("Ignoring $dist-$ver (due to CPAN+ bugs)");
+		$ignored = 1;
+      }
+
+      # check for previously parsed package string
+    } elsif ($line =~ /^(.*)\-(.+)$/) {
+      my $dist = $1;
+      my @dirs = split /\/+/, $dist;
+      my $ver  = $2;
+
+      if (@dirs == 1) {		# previously parsed
+		if (exists $paths{$dist}) {
+		  unshift @{ $paths{$dist} }, $ver;
+		} else {	
+		  $paths{$dist} = [ $ver ];
+		}
+      }
+    }
+  }
+  $self->_audit('')	if($ignored);
+  return %paths;
 }
 
 =head1 PROCEDURAL INTERFACE
@@ -534,7 +604,7 @@ sub test {
   my %config = ref($_[0]) eq 'HASH' ? %{ shift() } : ();
   $smoker ||= __PACKAGE__->new(%config);
 
-  $smoker->_audit("\n\n".('-'x40)."\n");
+  $smoker->_audit("\n".('-'x40)."\n");
 
   my @distros = @_;
   unless (@distros) {
@@ -559,61 +629,94 @@ sub test {
     my @dirs     = split /\/+/, $distpath;
     my $dist     = $dirs[-1];
 
-		# When there are multiple recent versions of a distribution, we
-		# only want to test the latest one. If it fails, then we'll
-		# check previous distributions.
+	# When there are multiple recent versions of a distribution, we
+	# only want to test the latest one. If it fails, then we'll
+	# check previous distributions.
 
-		my $passed     = 0;
-		my $fail_count = 0;
+    my $passed     = 0;
+    my $fail_count = 0;
+	my $report     = 1;
 
-		while ( (!$passed) && ($fail_count < $smoker->{fail_max}) &&
-				(my $ver = shift @versions) ) {
-			my $distpathver = join("-", $distpath, $ver);
-			my $distver     = join("-", $dist,     $ver);
+    # TODO - if test fails due to bad prereqs, set $fail_count to
+    # fail_max and abort testing versions (based on an option)
 
-			my $grade = $smoker->{checked}->{$distver}
-			  || 'ungraded';
+    while ( (!$passed) && ($fail_count < $smoker->{fail_max}) &&
+	    (my $ver = shift @versions) ) {
+      my $distpathver = join("-", $distpath, $ver);
+      my $distver     = join("-", $dist,     $ver);
 
-			if ((!defined $grade) ||
-			  $grade =~ /unknown|ungraded|none/) {
+      my $grade = $smoker->{checked}->{$distver} || 'ungraded';
 
-			  my $mod = $smoker->{cpan}->parse_module( module => $distpathver)
-			    or error("Invalid distribution $distver\n");
+      if ((!defined $grade) ||
+	      ($smoker->{allow_retries} && $grade =~ /$smoker->{allow_retries}/)) {
 
-				if ($mod && (!$mod->is_bundle)) {
-					$smoker->_audit("\n".('-'x40)."\n");
-					$smoker->_track("Testing $distpathver");
-					$smoker->{test_max}--;
+		# We no longer re-test 'unknown' since that rating is reserved
+		# for modules with no tests.
 
-					eval {
+	    my $mod = $smoker->{cpan}->parse_module( module => $distpathver)
+	      or error("Invalid distribution $distver\n");
 
-					  CPANPLUS::Error->flush();
+	    if ($mod && (!$mod->is_bundle)) {
+	      $smoker->_audit(('-'x40)."\n");
+	      $smoker->_track("Testing $distpathver");
+	      $smoker->{test_max}--;
+		  $report = 1;
 
-	my $stat = $smoker->{cpan}->install( 
-		modules  => [ $mod ],
-		target   => 'create',
-		allow_build_interactively => 0,
-		# other settings not set via set_confi() method
-	);
+		  eval {
+					  
+			CPANPLUS::Error->flush();
 
-	$smoker->_audit(CPANPLUS::Error->stack_as_string());
+			# TODO: option to not re-test prereqs that are known to
+			# pass (maybe if we use DBD::SQLite for the database and
+			# mark the date of the result?)
 
-	$smoker->{checked}->{$distver} = 'ungraded'	
-		unless ($grade = $smoker->{checked}->{$distver});
-	$passed = ($grade eq 'pass');
+			my $stat = $smoker->{cpan}->install( 
+				modules  => [ $mod ],
+				target   => 'create',
+				allow_build_interactively => 0,
+				# other settings now set via set_config() method
+			);
 
-	$smoker->_audit("\nReport Grade for $distver is ".uc($smoker->{checked}->{$distver})."\n");
+			# TODO: check the $stat and react appropriately
 
-					}; # end eval block
-				}
-			} else {
-				$passed = ($grade eq 'pass');
-				$smoker->_audit("$distpathver already tested and graded ".uc($grade)."\n");
-			}
-			$fail_count++, unless ($passed);
+			$smoker->_audit(CPANPLUS::Error->stack_as_string());
+
+			# TODO: option to mark uncompleted tests as aborted vs ungraded
+			#       aborted should indicate a fault in testing the distribution
+			#       ungraded should indicate a fault in testing a prerequisite
+			# 'Out of memory' faults, known failing prereqs, CPANPLUS faults,
+			# etc should all be covered by these. Otherwise it would be a FAIL.
+
+			$grade  = ($smoker->{checked}->{$distver} ||= 'aborted');
+			$passed = ($grade eq 'pass');
+
+			$smoker->_audit("\nReport Grade for $distver is ".uc($smoker->{checked}->{$distver})."\n");
+
+		  }; # end eval block
 		}
-	}
-	$smoker = undef;
+      } else {
+		if($report == 1) {
+		  $smoker->_audit(('-'x40)."\n");
+		  $report = 0;
+		}
+		$passed = ($grade eq 'pass');
+		$smoker->_audit("$distpathver already tested and graded ".uc($grade)."\n");
+      }
+      $fail_count++, unless ($passed);
+
+      # Mark older versions so that they are not tested
+      if ($passed) {
+		while (my $ver = shift @versions) {
+		  my $distver = join("-", $dist, $ver);
+		  $smoker->{checked}->{$distver} = "unknown";
+		}
+	  }
+    }
+  }
+  $smoker = undef;
+
+  # TODO: repository fills up. An option to flush it is needed.
+
 }
 
 =item mark( [ %config, ] $dist [, $grade ] ] )
@@ -630,6 +733,7 @@ make use of it.
 
 Grades can be one of (case insensitive):
 
+  aborted
   pass
   fail
   unknown
@@ -652,7 +756,7 @@ sub mark {
   my %config = ref($_[0]) eq 'HASH' ? %{ shift() } : ( verbose => 1, );
   $smoker ||= __PACKAGE__->new( );
 
-  $smoker->_audit("\n\n".('-'x40)."\n");
+  $smoker->_audit("\n".('-'x40)."\n");
 
   my $distver = shift || "";
   my $grade   = lc shift || "";
@@ -673,12 +777,12 @@ sub mark {
     );
     foreach my $dist (sort { versioncmp($a, $b) } keys %paths) {
       foreach my $ver (@{ $paths{$dist} }) {
-	$grade = $smoker->{checked}->{"$dist-$ver"};
-	if ($grade) {
-	  $smoker->_track("result for '$dist-$ver' is '$grade'");
-	} else {
-	  $smoker->_track("no result for '$dist-$ver'");
-	}
+		$grade = $smoker->{checked}->{"$dist-$ver"};
+		if ($grade) {
+		  $smoker->_track("result for '$dist-$ver' is '$grade'");
+		} else {
+		  $smoker->_track("no result for '$dist-$ver'");
+		}
       }
     }
   }
@@ -700,29 +804,167 @@ For further use of configuration settings see the new() constructor.
 =cut
 
 sub excluded {
-	my $smoker;
-	eval {
-		if ((ref $_[0]) && $_[0]->isa(__PACKAGE__)) {
-			$smoker = shift;
-		}
-	};
-	my %config = ref($_[0]) eq 'HASH' ? %{ shift() } : ();
-	$smoker ||= __PACKAGE__->new(%config);
+  my $smoker;
+  eval {
+    if ((ref $_[0]) && $_[0]->isa(__PACKAGE__)) {
+      $smoker = shift;
+    }
+  };
+  my %config = ref($_[0]) eq 'HASH' ? %{ shift() } : ();
+  $smoker ||= __PACKAGE__->new(%config);
 
-	$smoker->_audit("\n\n".('-'x40)."\n");
+  $smoker->_audit("\n".('-'x40)."\n");
 
-	my @distros = @_;
-	unless (@distros) {
-		@distros = $smoker->{plugin}->download_list();
-		unless (@distros) {
-			exit err("No new distributions uploaded to be tested");
+  my @distros = @_;
+  unless (@distros) {
+    @distros = $smoker->{plugin}->download_list();
+    unless (@distros) {
+      exit err("No new distributions uploaded to be tested");
+    }
+  }
+
+  my @dists = $smoker->_remove_excluded_dists( @distros );
+  $smoker->_audit('EXCLUDED: '.(scalar(@distros) - scalar(@dists))." distributions\n\n");
+  $smoker = undef;
+  return @dists;
+}
+
+# TODO: a method to purge older versions of test results from Checked
+# database. (That is, if the latest version tested is 1.23, we don't
+# need to keep earlier results around.)  There should be an option to
+# disable this behaviour.
+
+=item purge( [ %config, ] [ $dist [, $dist ... ] ] )
+
+  perl -MCPAN::YACSmoke -e purge()
+
+  perl -MCPAN::YACSmoke -e purge('Some-Dist-0.01')
+
+Purges the entries from the local cpansmoke database. The criteria for purging
+is that a distribution must have a more recent version, which has previously
+been marked as a PASS. However, if one or more distributions are passed as a
+parameter list, those specific distributions will be purged.
+
+If the flush_flag is set, via the config hash, to a true value, the directory 
+path created for each older copy of a distribution is deleted.
+
+For further use of configuration settings see the new() constructor.
+
+=cut
+
+sub purge {
+  my $smoker;
+  eval {
+    if ((ref $_[0]) && $_[0]->isa(__PACKAGE__)) {
+      $smoker = shift;
+    }
+  };
+  my %config = ref($_[0]) eq 'HASH' ? %{ shift() } : ();
+  $smoker ||= __PACKAGE__->new(%config);
+
+  my $flush = $smoker->{flush_flag} || 0;
+  my %distvars;
+  my $override = 0;
+
+  if(@_) {
+	  $override = 1;
+	  for(@_) {
+		next	unless(/^(.*)\-(.+)$/);
+		push @{$distvars{$1}}, $2;
+	  }
+  } else {
+	  for(keys %{$smoker->{checked}}) {
+		next	unless(/^(.*)\-(.+)$/);
+		push @{$distvars{$1}}, $2;
+	  }
+  }
+
+  for my $dist (sort keys %distvars) {
+	  my $passed = $override;
+	  my @vers = sort { versioncmp($a, $b) } @{$distvars{$dist}};
+	  while(@vers) {
+		my $vers = pop @vers;		# the latest
+		if($passed) {
+			$smoker->_track("'$dist-$vers' ['".
+							uc $smoker->{checked}->{"$dist-$vers"}.
+							"'] has been purged");
+			delete $smoker->{checked}->{"$dist-$vers"};
+			if($flush) {
+				my $builddir = catfile($smoker->basedir(), "$dist-$vers");
+				rmtree($builddir)	if(-d $builddir);
+			}
+			$smoker->_track("'$dist-$vers' ['".
+							uc $smoker->{checked}->{"$dist-$vers"}.
+							"'] has been purged");
 		}
+		elsif($smoker->{checked}->{"$dist-$vers"} eq 'pass') {
+			$passed = 1;
+		}
+	  }
+  }
+
+}
+
+=item flush( [ %config, ] [ 'all' | 'old' ]  )
+
+  perl -MCPAN::YACSmoke -e flush()
+  
+  perl -MCPAN::YACSmoke -e flush('all')
+
+  perl -MCPAN::YACSmoke -e flush('old')
+
+Removes unrequired build directories from the designated CPANPLUS build
+directory. Note that this deletes directories regardless of whether the 
+associated distribution was tested.
+
+Default flush is 'all'. The 'old' option will only delete the older 
+distributions, of multiple instances of a distribution.
+
+Note that this cannot be done reliably using last access or modify time, as
+the intention is for this distribution to be used on any OS that CPANPLUS
+is installed on. In this case not all OSs support the full range of return
+values from the stat function.
+
+For further use of configuration settings see the new() constructor.
+
+=cut
+
+sub flush {
+  my $smoker;
+  eval {
+    if ((ref $_[0]) && $_[0]->isa(__PACKAGE__)) {
+      $smoker = shift;
+    }
+  };
+  my %config = ref($_[0]) eq 'HASH' ? %{ shift() } : ();
+  $smoker ||= __PACKAGE__->new(%config);
+
+  my $param = shift || 'all';
+  my %dists;
+
+  opendir(DIR, $smoker->builddir());
+  while(my $dir = readdir(DIR)) {
+	next	if($dir =~ /^\.+$/);
+
+	if($param eq 'old') {
+		$dir =~ /(.*)-(.+)$extn/;
+		$dists{$1}->{$2} = "$dir";
+	} else {
+		rmtree($dir);
+		$smoker->_track("'$dir' flushed");
 	}
+  }
+  closedir(DIR);
+
+  if($param eq 'old') {
+	for my $dist (keys %dists) {
+	  for(sort { versioncmp($a, $b) } keys %{$dists{$dist}}) {
+	    rmtree($dists{$dist}->{$_});
+		$smoker->_track("'$dists{$dist}->{$_}' flushed");
+	  }
+	}
+  }
 
-	my @dists = $smoker->_remove_excluded_dists( @distros );
-	$smoker->_audit('EXCLUDED: '.(scalar(@distros) - scalar(@dists))." distributions\n\n");
-	$smoker = undef;
-	return @dists;
 }
 
 ## Private Methods
@@ -734,20 +976,20 @@ sub _track {
 }
 
 sub _debug {
-	my ($self,$message) = @_;
-	return	unless($self->{debug});
-	$self->_audit($message);
+  my ($self,$message) = @_;
+  return unless($self->{debug});
+  $self->_audit($message);
 }
 
 sub _audit {
-	my $self = shift;
-	$self->{audit_cb}->(@_)	if($self->{audit_cb});
-	return	unless($self->{audit_log});
+  my $self = shift;
+  $self->{audit_cb}->(@_)	if($self->{audit_cb});
+  return	unless($self->{audit_log});
 
-	my $FH = IO::File->new(">>".$self->{audit_log})
-		or exit error("Failed to write to file [$self->{audit_log}]: $!\n");
-	print $FH join("\n",@_) . "\n";
-	$FH->close;
+  my $FH = IO::File->new(">>".$self->{audit_log})
+    or exit error("Failed to write to file [$self->{audit_log}]: $!\n");
+  print $FH join("\n",@_) . "\n";
+  $FH->close;
 }
 
 1;
@@ -777,96 +1019,18 @@ Pass through configuration settings:
 	  recent_list_path => '.'
   };
 
-=head2 Outlook
+=head2 SmokeDB
 
-The Outlook plugin uses a mailbox via the Outlook mail client to return
-the current list of distributions
+The SmokeDB plugin uses the contents of the locally stored cpansmoke database.
+This can then be used to retest distributions that haven't been fully tested
+previously.
 
-Pass through configuration settings:
-
-  %config = {
-	  list_from => 'Outlook',
- 	  mailbox => 'CPAN Testers'
-  };
-
-Note that the mailbox needs to be in a folder below 'Inbox'.
-
-In order to use this plugin you will need to install the following
-modules:
-
-  Win32::OLE
-  Win32::OLE::Const
-
-=head2 NNTPWeb
-
-The NNTP plugin uses the webpages that are available at 
-http://www.nntp.perl.org/group/perl.cpan.testers/ to return
-the current list of distributions
-
-Pass through configuration settings:
-
-  %config = {
-	  list_from => 'NNTPWeb',
-	  nntp_id => $nntp_id
-  };
-
-In order to use this plugin you will need to install the following
-modules:
-
-  WWW::Mechanize
-  Template::Extract
-
-=head2 NNTP
-
-The NNTP plugin access the USENET group directly to return
-the current list of distributions
-
-Pass through configuration settings:
-
-  %config = {
-	  list_from => 'NNTP',
-	  nntp_id => $nntp_id
-  };
-
-In order to use this plugin you will need to install the following
-modules:
-
-  Net::NNTP
-
-=head2 WebList
-
-The WebList plugin accesses a formatted webpage to return
-the current list of distributions. The current pages which provide
-this list in correct format are cycled through, unless a specific
-webpage is requested.
-
-Currently used webpages:
-
-  http://cpan.uwinnipeg.ca/recent
-
-Note that the following webpages do not provide a 'download' link,
-which can identify the exact distribution. Further checking may be
-added in the future.
-
-  http://search.cpan.org/recent
-  http://use.perl.org/modulelist/
-
-Pass through configuration settings:
-
-  %config = {
-	  list_from => 'WebList',
- 	  webpath => 'http://cpan.uwinnipeg.ca/recent'
-  };
-
-In order to use this plugin you will need to install the following
-modules:
-
-  WWW::Mechanize
-  Template::Extract
+There are no pass through configuration settings.
 
 =head2 Writing A Plugin
 
-For an example, see one of the above plugins. 
+For an example, see one of the above plugins, or one of the known plugin
+packages available on CPAN (see list below). 
 
 The constructor, new(), is passed a hash of the configuration settings. The
 setting 'smoke' is an object reference to YACSmoke. Be sure to save the 
@@ -878,6 +1042,33 @@ that if a parameter value of 1 is passed to download_list(), this indicates
 that a test run is in progress, otherwise only a query on the outstanding 
 list is being made.
 
+=head2 Known Plugin Packages on CPAN
+
+=over
+
+=item CPAN::YACSmoke::Plugin::NNTP
+
+Uses the NNTP feed direct via 'nntp.perl.org' of 'perl.cpan.testers'.
+
+=item CPAN::YACSmoke::Plugin::NNTPWeb
+
+Uses the web interface to the newsgroup 'perl.cpan.testers'.
+
+=item CPAN::YACSmoke::Plugin::Outlook
+
+Uses a list of emails held in an Outlook mail folder
+
+=item CPAN::YACSmoke::Plugin::Phlanax100
+
+Uses the distributions within the Phlanax 100 list
+
+=item CPAN::YACSmoke::Plugin::WebList
+
+Uses a web page containing the list of recently uploaded distributions to
+CPAN. Uses KobeSearch by default, but another similar page can be requested.
+
+=back
+
 =head1 CAVEATS
 
 This is a proto-type release. Use with caution and supervision.
@@ -918,11 +1109,6 @@ L<http://www.missbarbell.co.uk>
 
 Jos Boumans <kane at cpan.org> for writing L<CPANPLUS>.
 
-=head2 Suggestions and Bug Reporting
-
-Please submit suggestions and report bugs to the CPAN Bug Tracker at
-L<http://rt.cpan.org>.
-
 =head1 COPYRIGHT AND LICENSE
 
 Copyright (C) 2005 by Robert Rothenberg.  All Rights Reserved.
